{{  Mini SD readonly Filesystem with FAT16/32 and SD/SDHC support
'   made by Andy Schenk (Ariba), July 2010
'
'   fsrw 2.6 Copyright 2009  Tomas Rokicki and Jonathan Dummer
'
'   See end of file for terms of use.
'
'   This object provides FAT16/32 file read/write access on a block device.
'   Only one file open at a time.  Open modes are 'r' (read), 'a' (append),
'   'w' (write), and 'd' (delete).  Only the root directory is supported.
'   No long filenames are supported.  We also support traversing the
'   root directory.
'
'   In general, negative return values are errors; positive return
'   values are success.  Other than -1 on popen when the file does not
'   exist, all negative return values will be "aborted" rather than
'   returned.
'
'   Changes:
'       v1.1  28 December 2006  Fixed offset for ctime
'       v1.2  29 December 2006  Made default block driver be fast one
'       v1.3  6 January 2007    Added some docs, and a faster asm
'       v1.4  4 February 2007   Rearranged vars to save memory;
'                               eliminated need for adjacent pins;
'                               reduced idle current consumption; added
'                               sample code with abort code data
'       v1.5  7 April 2007      Fixed problem when directory is larger
'                               than a cluster.
'       v1.6  23 September 2008 Fixed a bug found when mixing pputc
'                               with pwrite.  Also made the assembly
'                               routines a bit more cautious.
'       v2.1  12 July 2009      FAT32, SDHC, multiblock, bug fixes
'       v2.4  26 September 2009 Added seek support.  Added clustersize.
'       v2.4a   6 October 2009 modified setdate to explicitly set year/month/etc.
'       v2.5  13 November 2009 fixed a bug on releasing the pins, added a "release" pass through function
'       v2.6  11 December 2009: faster transfer hub <=> cog, safe_spi.spin uses 1/2 speed reads, is default
'       mini version July 2010
}}
'
'   Constants describing FAT volumes.
'
CON

   SECTORSIZE = 512
   SECTORSHIFT = 9
   DIRSIZE = 32
   DIRSHIFT = 5

   BOOTMEM  = $1800          'Boot buffer begin  (max 26kByte files)
   SD_BASEPIN = 0
'
'   The object that provides the block-level access.
'
OBJ

   sdspi: "sdspi_mini_rd2"
   
VAR

'
'   Variables concerning the open file.
'
   long fclust ' the current cluster number
   long filesize ' the total current size of the file
   long floc ' the seek position of the file
   long frem ' how many bytes remain in this cluster from this file
   long bufat ' where in the buffer our current character is
   long bufend ' the last valid character (read) or free position (write)
   long direntry ' the byte address of the directory entry (if open for write)
   long writelink ' the byte offset of the disk location to store a new cluster
   long fatptr ' the byte address of the most recently written fat entry
   long firstcluster ' the first cluster of this file

'   Variables used when mounting to describe the FAT layout of the card
'   (moved to the end of the file in the Spin version).
'
'
'   Variables controlling the caching.
'
'
'  Buffering:  two sector buffers.  These two buffers must be longword
'  aligned!  To ensure this, make sure they are the first byte variables
'  defined in this object.
'
'   byte buf[SECTORSIZE]  ' main data buffer
'   byte buf2[SECTORSIZE] 'data buffer

   long buf, buf2        'pointers for two sector buffers (reused sdspi pasm hub data)

{PUB mount(basepin) : r | start, sectorspercluster, reserved, rootentries, sectors

   return mount_explicit(basepin, basepin+1, basepin+2, basepin+3)}
   
PUB mount_explicit(DO, CLK, DI, CS) : r | start, sectorspercluster, reserved, rootentries, sectors

{{
'   Mount a volume.  The address passed in is passed along to the block
'   layer; see the currently used block layer for documentation.  If the
'   volume mounts, a 0 is returned, else abort is called.
}}

   if (pdate == 0)
      pdate := constant(((2009-1980) << 25) + (1 << 21) + (27 << 16) + (7 << 11))
   unmount
   r := sdspi.start_explicit(DO, CLK, DI, CS)
   'if r < 0
     'abort r
   buf := r                     'set pointers to sector buffers
   buf2 := r+512
   lastread := -1
   dirty := 0
   sdspi.readblock(0, buf)
   if (getfstype > 0)
     start := 0
   else
     start := brlong(buf+$1c6)
     sdspi.readblock(start, buf)
   filesystem := getfstype
'   if (filesystem == 0)
'      abort(-20) ' not a fat16 or fat32 volume
'   if (brword(@buf+$0b) <> SECTORSIZE)
'      abort(-21) ' bad bytes per sector
   sectorspercluster := byte[buf+$0d]
'   if (sectorspercluster & (sectorspercluster - 1))
'      abort(-22) ' bad sectors per cluster
   clustershift := 0
   repeat while (sectorspercluster > 1)
      clustershift++
      sectorspercluster >>= 1
   sectorspercluster := 1 << clustershift
   clustersize := SECTORSIZE << clustershift
   reserved := brword(buf+$0e)
'   if (buf[$10] <> 2)
'      abort(-23) ' not two FATs
   sectors := brword(buf+$13)
   if (sectors == 0)
     sectors := brlong(buf+$20)
   fat1 := start + reserved
   if (filesystem == 2)
      rootentries := 16 << clustershift
      sectorsperfat := brlong(buf+$24)
      dataregion := (fat1 + 2 * sectorsperfat) - 2 * sectorspercluster
      rootdir := (dataregion + (brword(buf+$2c) << clustershift)) << SECTORSHIFT
      rootdirend := rootdir + (rootentries << DIRSHIFT)
      endofchain := $ffffff0
   else
      rootentries := brword(buf+$11)
      sectorsperfat := brword(buf+$16)
      rootdir := (fat1 + 2 * sectorsperfat) << SECTORSHIFT
      rootdirend := rootdir + (rootentries << DIRSHIFT)
      dataregion := 1 + ((rootdirend - 1) >> SECTORSHIFT) - 2 * sectorspercluster
      endofchain := $fff0
'   if (brword(@buf+$1fe) <> $aa55)
'     abort(-24) ' bad FAT signature
   totclusters := ((sectors - dataregion + start) >> clustershift)
   ' return r (default return)
'
'   For compatibility, a single pin.
'

PUB unmount

  pclose
  sdspi.stop
  
PUB release

{{   This is just a pass-through function to allow the block layer
     to tristate the I/O pins to the card.
}}

   sdspi.release

PUB opendir | off

{{
'   Close the currently open file, and set up the read buffer for
'   calls to nextfile.
}}

   pclose
   off := rootdir - (dataregion << SECTORSHIFT)
   fclust := off >> (clustershift + SECTORSHIFT)
   floc := off - (fclust << (clustershift + SECTORSHIFT))
   frem := rootdirend - rootdir
   filesize := floc + frem
   return 0
   
{PUB nextfile(fbuf) | i, t, at, lns

{{
'   Find the next file in the root directory and extract its
'   (8.3) name into fbuf.  Fbuf must be sized to hold at least
'   13 characters (8 + 1 + 3 + 1).  If there is no next file,
'   -1 will be returned.  If there is, 0 will be returned.
}}

   repeat
      if (bufat => bufend)
         t := pfillbuf
         if (t < 0)
            return t
         if (((floc >> SECTORSHIFT) & ((1 << clustershift) - 1)) == 0)
            fclust++
      at := buf + bufat
      if (byte[at] == 0)
         return -1
      bufat += DIRSIZE
      if (byte[at] <> $e5 and (byte[at][$0b] & $18) == 0)
         lns := fbuf
         repeat i from 0 to 10
            byte[fbuf] := byte[at][i]
            fbuf++
            if (byte[at][i] <> " ")
               lns := fbuf
            if (i == 7 or i == 10)
               fbuf := lns
               if (i == 7)
                  byte[fbuf] := "."
                  fbuf++
         byte[fbuf] := 0
         return 0 }

PUB popen(s, mode) : r | i, sentinel, dirptr, freeentry

{{
'   Close any currently open file, and open a new one with the given
'   file name and mode.  Mode can be "r" "w" "a" or "d" (delete).
'   If the file is opened successfully, 0 will be returned.  If the
'   file did not exist, and the mode was not "w" or "a", -1 will be
'   returned.  Otherwise abort will be called with a negative error
'   code.
}}

   pclose
   i := 0
   repeat while (i<8 and byte[s] and byte[s] <> ".")
      padname[i++] := uc(byte[s++])
   repeat while (i<8)
      padname[i++] := " "
   repeat while (byte[s] and byte[s] <> ".")
      s++
   if (byte[s] == ".")
      s++
   repeat while (i<11 and byte[s])
      padname[i++] := uc(byte[s++])
   repeat while (i < 11)
      padname[i++] := " "
   sentinel := 0
   freeentry := 0
   repeat dirptr from rootdir to rootdirend - DIRSIZE step DIRSIZE
      s := readbytec(dirptr)
      if (freeentry == 0 and (byte[s] == 0 or byte[s] == $e5))
         freeentry := dirptr
      if (byte[s] == 0)
         sentinel := dirptr
         quit
      repeat i from 0 to 10
         if (padname[i] <> byte[s][i])
            quit
      if (i == 11 and 0 == (byte[s][$0b] & $18)) ' this always returns
         fclust := brword(s+$1a)
         if (filesystem == 2)
            fclust += brword(s+$14) << 16
         firstcluster := fclust
         filesize := brlong(s+$1c)
         if (mode == "r")
            frem := (clustersize) <# (filesize)
            return 0
   return -1   ' not found

PUB pclose : r

{{
'   Flush and close the currently open file if any.  Also reset the
'   pointers to valid values.  If there is no error, 0 will be returned.
}}

   bufat := 0
   bufend := 0
   filesize := 0
   floc := 0
   frem := 0
   writelink := 0
   direntry := 0
   fclust := 0
   firstcluster := 0
   sdspi.release
   ' return r (default return)

PUB get_filesize

   return filesize

PUB pread(ubuf, count) : r | t

{{
'   Read count bytes into the buffer ubuf.  Returns the number of bytes
'   successfully read, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}

   repeat while (count > 0)
      if (bufat => bufend)
         t := pfillbuf
         if (t =< 0)
            if (r > 0)
' parens below prevent this from being optimized out
              return (r)
            return t
      t := (bufend - bufat) <# (count)
      if ((t | (ubuf) | bufat) & 3)
         bytemove(ubuf, buf+bufat, t)
      else
         longmove(ubuf, buf+bufat, t>>2)
      bufat += t
      r += t
      ubuf += t
      count -= t
   ' return r (default return)

PRI readblockc(n)

'   Read a block into the metadata buffer, if that block is not already
'   there.

   if (n <> lastread)
      sdspi.readblock(n, buf2)
      lastread := n

PRI brword(b)

'   Read a byte-reversed word from a (possibly odd) address.

   return (byte[b]) + ((byte[b][1]) << 8)
   
PRI brlong(b)

'   Read a byte-reversed long from a (possibly odd) address.

   return brword(b) + (brword(b+2) << 16)
   
PRI brclust(b)

'   Read a cluster entry.

  if (filesystem == 1)
    return brword(b)
  else
    return brlong(b)
    
PRI getfstype : r

  if (brlong(buf+$36) == constant("F" + ("A" << 8) + ("T" << 16) + ("1" << 24)) and byte[buf+$3a]=="6")
    return 1
  if (brlong(buf+$52) == constant("F" + ("A" << 8) + ("T" << 16) + ("3" << 24)) and byte[buf+$56]=="2")
    return 2
  ' return r (default return)
   
PRI readbytec(byteloc)

'   Read a byte address from the disk through the metadata buffer and
'   return a pointer to that location.

   readblockc(byteloc >> SECTORSHIFT)
   return buf2 + (byteloc & constant(SECTORSIZE - 1))
   
PRI readfat(clust)

'   Read a fat location and return a pointer to the location of that
'   entry.

   fatptr := (fat1 << SECTORSHIFT) + (clust << filesystem)
   return readbytec(fatptr)
   
PRI followchain : r

'   Follow the fat chain and update the writelink.

   r := brclust(readfat(fclust))
   writelink := fatptr
   ' return r (default return)
   
PRI nextcluster : r

'   Read the next cluster and return it.  Set up writelink to
'   point to the cluster we just read, for later updating.  If the
'   cluster number is bad, return a negative number.

   r := followchain
   if (r < 2 or r => totclusters)
      abort(-9) ' bad cluster value
   ' return r (default return)

PRI datablock

'   Calculate the block address of the current data location.

   return (fclust << clustershift) + dataregion + ((floc >> SECTORSHIFT) & ((1 << clustershift) - 1))
   
PRI uc(c)

'   Compute the upper case version of a character.

   if ("a" =< c and c =< "z")
      return c - 32
   return c

PRI pfillbuf : r

'   Get some data into an empty buffer.  If no more data is available,
'   return -1.  Otherwise return the number of bytes read into the
'   buffer.

   if (floc => filesize)
      return -1
   if (frem == 0)
      fclust := nextcluster
      frem := (clustersize) <# (filesize - floc)
   sdspi.readblock(datablock, buf)
   r := SECTORSIZE
   if (floc + r => filesize)
      r := filesize - floc
   floc += r
   frem -= r
   bufat := 0
   bufend := r
   ' return r (default return)
   
DAT

filesystem    long      0 ' 0 = unmounted, 1 = fat16, 2 = fat32
rootdir       long      0 ' the byte address of the start of the root directory
rootdirend    long      0 ' the byte immediately following the root directory.
dataregion    long      0 ' the start of the data region, offset by two sectors
clustershift  long      0 ' log base 2 of blocks per cluster
clustersize   long      0 ' total size of cluster in bytes
fat1          long      0 ' the block address of the fat1 space
totclusters   long      0 ' how many clusters in the volume
sectorsperfat long      0 ' how many sectors per fat
endofchain    long      0 ' end of chain marker (with a 0 at the end)
pdate         long      0 ' current date
lastread      long      0 ' the block address of the buf2 contents
dirty         long      0 ' nonzero if buf2 is dirty
padname       byte      0[11]  ' filename buffer